import os
import shutil
from ..utils.config import get_module_filepath

def init_project(args):
    """
    Initializes a new sunholo project with a basic configuration file and directory structure.

**Template Files (`templates/project`):**

A `templates/project` directory is within the `sunholo` package with the following template files in it:

* **`config/llm_config.yaml`:** A basic configuration file with placeholders for LLM settings, vector stores, etc.
* **`config/cloud_run_urls.json`:** A template for Cloud Run URLs.
* **`app.py`:** A basic Flask app that can be customized for the project.
* **`.gitignore`:** A gitignore file to exclude unnecessary files from version control.
* **`README.md`:** A README file with instructions for setting up and running the project.

**Usage:**

Users can initialize a new project using the following command:

```bash
sunholo init my_genai_project
```

This will create a new directory named `my_genai_project` with the template files, allowing users to start building their GenAI application.
    """
    project_name = args.project_name
    project_dir = os.path.join(os.getcwd(), project_name)

    print(f"Initializing project: {project_name} in directory: {project_dir}")

    # Create project directory
    if os.path.exists(project_dir):
        print(f"Directory {project_dir} already exists. Please choose a different project name.")
        return

    os.makedirs(project_dir)

    # Copy template files
    template_dir = get_module_filepath("templates/project")
    for filename in os.listdir(template_dir):
        src_path = os.path.join(template_dir, filename)
        dest_path = os.path.join(project_dir, filename)
        if os.path.isfile(src_path):
            shutil.copy(src_path, dest_path)
        elif os.path.isdir(src_path):
            shutil.copytree(src_path, dest_path)

    print(f"Project {project_name} initialized successfully.")
    print(f"Navigate to {project_dir} and customize the configuration files in the 'config' directory.")

def setup_init_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'init' command.
    """
    init_parser = subparsers.add_parser('init', help='Initializes a new Multivac project.')
    init_parser.add_argument('project_name', help='The name of the new project.')
    init_parser.set_defaults(func=init_project)



from .discovery_engine_client import DiscoveryEngineClient
from ..utils.config import load_config_key
from ..utils.gcp_project import get_gcp_project

def create_new_discovery_engine(vector_name):

    chunker_config = load_config_key("chunker", vector_name=vector_name, kind="vacConfig")

    chunk_size = 500
    if chunker_config:
        if "chunk_size" in chunker_config:
            chunk_size = chunker_config["chunk_size"]    

    project_id = get_gcp_project()
    if not project_id:
        raise ValueError("Could not find project_id in gcp_config")
    
    #location = gcp_config.get('location')

    de = DiscoveryEngineClient(
                    data_store_id=vector_name, 
                    project_id=project_id,
                    # location needs to be 'eu' or 'us' which doesn't work with other configurations
                    #location=location
                    )

    new_store = de.create_data_store(chunk_size=chunk_size)

    return new_store


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Any, Dict, List, Union
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.schema import LLMResult

import threading
import re
from ..logging import log

class ContentBuffer:
    """
    A buffer class for storing and managing textual content.

    This class provides methods to write text to the buffer, read the entire buffer content,
    and clear the buffer content. The buffer can be used to collect text output for further
    processing or inspection.

    Attributes:
        content (str): Stores the textual content of the buffer.
    """

    def __init__(self):
        """
        Initializes a new ContentBuffer instance.

        The content buffer starts with an empty string, and logging is initialized to indicate
        that the buffer has been created.
        """
        self.content = ""
        log.debug("Content buffer initialized")
    
    def write(self, text: str):
        """
        Writes text to the content buffer.

        Args:
            text (str): The text to be added to the buffer.

        Adds the given text to the existing content of the buffer.
        """
        self.content += text
    
    def read(self) -> str:
        """
        Reads the entire content from the buffer.

        Returns:
            str: The content of the buffer.

        Provides the entire content stored in the buffer.
        """   
        return self.content

    def clear(self):
        """
        Clears the content buffer.

        Empties the buffer content, resetting it to an empty string.
        """
        self.content = ""

    
class BufferStreamingStdOutCallbackHandler(StreamingStdOutCallbackHandler):
    """
    A callback handler for streaming LLM output to a content buffer.

    This class handles the streaming of output from a large language model (LLM),
    processes tokens from the model output, and writes them to a ContentBuffer.
    It supports handling different types of tokens and keeps track of code blocks
    and questions.

    Attributes:
        content_buffer (ContentBuffer): The buffer to which content is streamed.
        tokens (str): Tokens that indicate the end of a statement, for buffer flushing.
        buffer (str): Temporary storage for accumulating streamed tokens.
        stream_finished (threading.Event): Signals when the streaming is finished.
        in_code_block (bool): Indicates whether the current context is a code block.
        in_question_block (bool): Indicates whether the current context is a question block.
        question_buffer (str): Stores the accumulated questions.
    """

    def __init__(self, content_buffer: ContentBuffer, tokens: str = ".?!\n", *args, **kwargs):
        """
        Initializes a new BufferStreamingStdOutCallbackHandler instance.

        Args:
            content_buffer (ContentBuffer): The buffer to which content will be written.
            tokens (str): Tokens that indicate the end of a statement (default: ".?!\n").
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Sets up the callback handler with the given content buffer and tokens.
        Initializes tracking variables for code blocks, buffer content, and the finished signal.
        """
        super().__init__(*args, **kwargs)
        self.content_buffer = content_buffer
        self.tokens = tokens
        self.buffer = ""
        self.stream_finished = threading.Event()
        self.in_code_block = False
        self.in_question_block = False
        self.question_buffer = ""
        log.info("Starting to stream LLM")

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """
        Processes a new token from the LLM output.

        Args:
            token (str): The new token generated by the LLM.
            **kwargs: Additional keyword arguments.

        Accumulates the token in the buffer and processes it based on the current context.
        The buffer content is written to the content buffer when appropriate tokens or
        patterns are detected.
        """
        log.debug(f"on_llm_new_token: {token}")
        
        self.buffer += token

        if '```' in token:
            self.in_code_block = not self.in_code_block

        if not self.in_code_block:
            self._process_buffer()

    def _process_buffer(self):
        """
        Processes the buffer content and writes to the content buffer.

        If the buffer ends with a numbered list pattern or specified tokens, the buffer is flushed
        to the content buffer. Otherwise, the buffer is left intact for further accumulation.
        """
        matches = list(re.finditer(r'\n(\d+\.\s)', self.buffer))
        if matches:
            last_match = matches[-1]
            start_of_last_match = last_match.start() + 1
            self.content_buffer.write(self.buffer[:start_of_last_match])
            self.buffer = self.buffer[start_of_last_match:]
        else:
            if any(self.buffer.endswith(t) for t in self.tokens):
                self.content_buffer.write(self.buffer)
                self.buffer = ""

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """
        Handles the end of LLM streaming.

        Args:
            response (LLMResult): The result returned by the LLM.
            **kwargs: Additional keyword arguments.

        Writes any remaining buffer content to the content buffer, and sets a signal indicating
        that the streaming has finished.
        """
        if self.buffer:
            self.content_buffer.write(self.buffer)
            self.buffer = ""
            log.info("Flushing remaining LLM response buffer")

        self.stream_finished.set()
        log.info("Streaming LLM response ended successfully")



try:
    from vertexai.preview import rag
    from vertexai.preview.generative_models import Tool, grounding, GenerationResponse
except ImportError:
    rag = None

from ..logging import log
from ..utils.config import load_config_key
from ..components import load_memories
from ..llamaindex.get_files import fetch_corpus
from ..discovery_engine.discovery_engine_client import DiscoveryEngineClient
from ..utils.gcp_project import get_gcp_project

def get_vertex_memories(vector_name):
    """
    Retrieves a LlamaIndex corpus from Vertex AI based on the provided Google Cloud configuration.
    
    This function constructs a corpus name using project details from the configuration and attempts
    to fetch the corresponding corpus. If the corpus cannot be retrieved, it raises an error.
    
    Parameters:
    - vector_name: The name of the of VAC
    
    Returns:
    - List of corpus objects fetched from Vertex AI.
    
    Raises:
    - ValueError: If any of the required configurations (project_id, location, or rag_id) are missing,
      or if the corpus cannot be retrieved.

    Example:
    ```python

    # Fetch the corpus
    try:
        corpus = get_corpus("edmonbrain")
        print("Corpus fetched successfully:", corpus)
    except ValueError as e:
        print("Error fetching corpus:", str(e))
    ```
    """
    gcp_config = load_config_key("gcp_config", vector_name=vector_name, kind="vacConfig")

    if not rag:
        raise ValueError("Need to install vertexai module via `pip install sunholo[gcp]`")
    
    global_location = gcp_config.get('location')

    memories = load_memories(vector_name)
    tools = []

    if not memories:
        return tools
    
    for memory in memories:
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore == "llamaindex":
                log.info(f"Found vectorstore {vectorstore}")
                rag_id = value.get('rag_id')
                if rag_id is None:
                    raise ValueError("Must specify rag_id if using vectorstore: llamaindex")
                
                project_id = value.get('project_id') or gcp_config.get('project_id')
                location = value.get('location') or gcp_config.get('location')

                try:
                    corpus = fetch_corpus(
                        project_id=project_id or get_gcp_project(),
                        location=location or global_location,
                        rag_id=rag_id
                    )
                    corpus_tool = Tool.from_retrieval(
                        retrieval=rag.Retrieval(
                            source=rag.VertexRagStore(
                                rag_corpora=[corpus.name],  # Currently only 1 corpus is allowed.
                                similarity_top_k=10,  # Optional
                            ),
                        )
                    )
                    tools.append(corpus_tool)
                except Exception as err:
                    log.error(f"Failed to fetch vertex.rag: {str(err)} - skipping")
                    continue

            elif vectorstore == "discovery_engine" or vectorstore == "vertex_ai_search":

                try:
                    project_id = value.get('project_id') or get_gcp_project()
                    if value.get('chunks'):
                        log.warning("Data stores for chunks do not work with Tools yet, call data store directly instead")
                        continue

                    if value.get('read_only'):
                        new_vector_name = value.get('vector_name')
                        if not new_vector_name:
                            log.warning("read_only specified but no new vector_name to read from")
                        vector_name = new_vector_name

                    de = DiscoveryEngineClient(vector_name, project_id=project_id)
                    log.info(f"Found vectorstore {vectorstore}")

                    data_store_path = f"{de.data_store_path()}/dataStores/{vector_name}"
                    corpus_tool = Tool.from_retrieval(
                        grounding.Retrieval(grounding.VertexAISearch(datastore=data_store_path))
                    )
                    tools.append(corpus_tool)
                except Exception as err:
                    log.error(f"Failed to fetch DiscoveryEngine grounding - {str(err)} - skipping")
                    continue
                

    if not tools:
        log.warning("No Vertex corpus configurations could be found")
    
    return tools

def get_google_search_grounding(vector_name):
    # can't have this and llamaindex memories?
    ground = load_config_key("grounding", vector_name=vector_name, kind="vacConfig")
    if ground and ground.get("google_search"):
        gs_tool = Tool.from_google_search_retrieval(grounding.GoogleSearchRetrieval())
        log.info(f"Got Search Tool: {gs_tool}")
        return gs_tool
    
    log.info(f"No google search config available for {vector_name}")
    return None

def print_grounding_response(response):
    """Prints Gemini response with grounding citations."""
    grounding_metadata = response.candidates[0].grounding_metadata

    # Citation indices are in byte units
    ENCODING = "utf-8"
    text_bytes = response.text.encode(ENCODING)

    prev_index = 0
    markdown_text = ""

    sources: dict[str, str] = {}
    footnote = 1
    for attribution in grounding_metadata.grounding_attributions:
        context = attribution.web or attribution.retrieved_context
        if not context:
            log.info(f"Skipping Grounding Attribution {attribution}")
            continue

        title = context.title
        uri = context.uri
        end_index = int(attribution.segment.end_index)

        if uri not in sources:
            sources[uri] = {"title": title, "footnote": footnote}
            footnote += 1

        text_segment = text_bytes[prev_index:end_index].decode(ENCODING)
        markdown_text += f"{text_segment} [[{sources[uri]['footnote']}]]({uri})"
        prev_index = end_index

    if prev_index < len(text_bytes):
        markdown_text += str(text_bytes[prev_index:], encoding=ENCODING)

    markdown_text += "\n## Grounding Sources\n"

    if grounding_metadata.web_search_queries:
        markdown_text += (
            f"\n**Web Search Queries:** {grounding_metadata.web_search_queries}\n"
        )
    elif grounding_metadata.retrieval_queries:
        markdown_text += (
            f"\n**Retrieval Queries:** {grounding_metadata.retrieval_queries}\n"
        )

    for uri, source in sources.items():
        markdown_text += f"{source['footnote']}. [{source['title']}]({uri})\n"
    
    log.info(markdown_text)
    return markdown_text


import json
from ..logging import log

def extract_chat_history(chat_history=None):
    """
    Extracts paired chat history between human and AI messages.

    This function takes a chat history and returns a list of pairs of messages,
    where each pair consists of a human message followed by the corresponding AI response.

    Args:
        chat_history (list): List of chat messages.

    Returns:
        list: List of tuples with paired human and AI messages.

    Example:
    ```python
    chat_history = [
        {"name": "Human", "text": "Hello, AI!"},
        {"name": "AI", "text": "Hello, Human! How can I help you today?"}
    ]
    paired_messages = extract_chat_history(chat_history)
    print(paired_messages)
    # Output: [("Hello, AI!", "Hello, Human! How can I help you today?")]
    ```
    """
    if not chat_history:
        log.info("No chat history found")
        return []

    log.info(f"Extracting chat history: {chat_history}")
    paired_messages = []

    first_message = chat_history[0]
    log.info(f"Extracting first_message: {first_message}")
    if is_bot(first_message):
        blank_human_message = {"name": "Human", "content": "", "embeds": []}
        paired_messages.append((create_message_element(blank_human_message), 
                                create_message_element(first_message)))
        chat_history = chat_history[1:]

    last_human_message = ""
    for message in chat_history:
        log.info(f"Extracing message: {message}")
        if is_human(message):
            last_human_message = create_message_element(message)
            log.info(f"Extracted human message: {last_human_message}")
        elif is_bot(message):
            ai_message = create_message_element(message)
            log.info(f"Extracted AI message: {ai_message}")
            paired_messages.append((last_human_message, ai_message))
            last_human_message = ""

    log.info(f"Paired messages: {paired_messages}")

    return paired_messages

def embeds_to_json(message: dict):
    """
    Converts the 'embeds' field in a message to a JSON string.

    Args:
        message (dict): The message containing the 'embeds' field.

    Returns:
        str: JSON string representation of the 'embeds' field or an empty string if no embeds are found.

    Example:
    ```python
    message = {"embeds": [{"type": "image", "url": "https://example.com/image.png"}]}
    json_string = embeds_to_json(message)
    print(json_string)
    # Output: '[{"type": "image", "url": "https://example.com/image.png"}]'
    ```
    """
    if 'embeds' in message and len(message['embeds']) > 0:
        return json.dumps(message.get("embeds"))
    else:
        return ""

def create_message_element(message: dict):
    """
    Extracts the main content of a message.

    Args:
        message (dict): The message to extract content from.

    Returns:
        str: The text or content of the message.

    Raises:
        KeyError: If neither 'content' nor 'text' fields are found.

    Example:
    ```python
    message = {"text": "Hello, AI!"}
    content = create_message_element(message)
    print(content)
    # Output: 'Hello, AI!'
    ```
    """
    if 'text' in message:  # This is a Slack or Google Chat message
        log.info(f"Found text element - {message['text']}")
        return message['text']
    elif 'content' in message: # Discord or OpenAI history message
        log.info(f"Found content element - {message['content']}")
        return message['content']
    else:  
        raise KeyError(f"Could not extract 'content' or 'text' element from message: {message}, {type(message)}")

def is_human(message: dict):
    """
    Checks if a message was sent by a human.

    Args:
        message (dict): The message to check.

    Returns:
        bool: True if the message was sent by a human, otherwise False.

    Example:
    ```python
    message = {"name": "Human"}
    print(is_human(message))
    # Output: True
    ```
    """
    if 'name' in message:
        return message["name"] == "Human"
    elif 'sender' in message:  # Google Chat
        return message['sender']['type'] == 'HUMAN'
    elif 'role' in message:
        return message['role'] == 'user'
    else:
        # Slack: Check for the 'user' field and absence of 'bot_id' field
        return 'user' in message and 'bot_id' not in message

def is_bot(message: dict):
    """
    Checks if a message was sent by a bot.

    Args:
        message (dict): The message to check.

    Returns:
        bool: True if the message was sent by a bot, otherwise False.

    Example:
    ```python
    message = {"name": "AI"}
    print(is_bot(message))
    # Output: True
    ```
    """
    return not is_human(message)

def is_ai(message: dict):
    """
    Checks if a message was specifically sent by an AI.

    Args:
        message (dict): The message to check.

    Returns:
        bool: True if the message was sent by an AI, otherwise False.

    Example:
    ```python
    message = {"name": "AI"}
    print(is_ai(message))
    # Output: True
    ```
    """
    if 'name' in message:
        return message["name"] == "AI"
    elif 'sender' in message:  # Google Chat
        return message['sender']['type'] == 'BOT'
    elif 'role' in message:
        return message['role'] == 'assistant'
    else:
        return 'bot_id' in message  # Slack



import os
import base64
import json
from datetime import datetime
try:
    from playwright.sync_api import sync_playwright
except ImportError:
    sync_playwright = None

class BrowseWebWithImagePromptsBot:
    """
    Examples:

    ```python
    class ProductionBot(BrowseWebWithImagePromptsBot):
        def send_prompt_to_llm(self, prompt, screenshot_base64):
            # Implement the actual logic to send the prompt and screenshot to the LLM and return the response
            api_url = "https://api.example.com/process"  # Replace with the actual LLM API endpoint
            headers = {"Content-Type": "application/json"}
            data = {
                "prompt": prompt,
                "screenshot": screenshot_base64
            }
            response = requests.post(api_url, headers=headers, data=json.dumps(data))
            return response.text  # Assuming the response is in JSON format

    @app.route('/run-bot', methods=['POST'])
    def run_bot():
        data = request.json
        session_id = data.get('session_id')
        website_name = data.get('website_name')
        browser_type = data.get('browser_type', 'chromium')
        current_action_description = data.get('current_action_description', "")
        next_goal = data.get('next_goal', "")
        
        bot = ProductionBot(session_id=session_id, website_name=website_name, browser_type=browser_type, headless=True)
        
        # Check if initial instructions are provided
        initial_instructions = data.get('instructions')
        if initial_instructions:
            bot.execute_instructions(initial_instructions)
        
        # Take initial screenshot and send to LLM if no instructions provided
        if not initial_instructions:
            screenshot_path = bot.take_screenshot()
            new_instructions = bot.send_screenshot_to_llm(screenshot_path, current_action_description, next_goal)
            bot.execute_instructions(new_instructions)
        
        # Take final screenshot
        bot.take_screenshot()
        
        bot.close()
        
        return jsonify({"status": "completed", "new_instructions": new_instructions})

    if __name__ == "__main__":
        app.run(host='0.0.0.0', port=8080)
    ```
    """
    def __init__(self, session_id, website_name, browser_type='chromium', headless=True):
        if not sync_playwright:
            raise ImportError("playright needed for BrowseWebWithImagePromptsBot class - install via `pip install sunholo[tools]`")
        self.session_id = session_id
        self.website_name = website_name
        self.browser_type = browser_type
        self.screenshot_dir = f"{website_name}_{session_id}"
        os.makedirs(self.screenshot_dir, exist_ok=True)
        self.cookie_file = os.path.join(self.screenshot_dir, "cookies.json")
        self.playwright = sync_playwright().start()
        
        if browser_type == 'chromium':
            self.browser = self.playwright.chromium.launch(headless=headless)
        elif browser_type == 'firefox':
            self.browser = self.playwright.firefox.launch(headless=headless)
        elif browser_type == 'webkit':
            self.browser = self.playwright.webkit.launch(headless=headless)
        else:
            raise ValueError(f"Unsupported browser type: {browser_type}")
        
        self.context = self.browser.new_context()
        self.page = self.context.new_page()
        self.load_cookies()

    def load_cookies(self):
        if os.path.exists(self.cookie_file):
            with open(self.cookie_file, 'r') as f:
                cookies = json.load(f)
                self.context.add_cookies(cookies)

    def save_cookies(self):
        cookies = self.context.cookies()
        with open(self.cookie_file, 'w') as f:
            json.dump(cookies, f)

    def navigate(self, url):
        self.page.goto(url)

    def click(self, selector):
        self.page.click(selector)

    def scroll(self, direction='down', amount=1):
        for _ in range(amount):
            if direction == 'down':
                self.page.evaluate("window.scrollBy(0, window.innerHeight)")
            elif direction == 'up':
                self.page.evaluate("window.scrollBy(0, -window.innerHeight)")
            elif direction == 'left':
                self.page.evaluate("window.scrollBy(-window.innerWidth, 0)")
            elif direction == 'right':
                self.page.evaluate("window.scrollBy(window.innerWidth, 0)")

    def type_text(self, selector, text):
        self.page.fill(selector, text)

    def take_screenshot(self):
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        screenshot_path = os.path.join(self.screenshot_dir, f"screenshot_{timestamp}.png")
        self.page.screenshot(path=screenshot_path)
        return screenshot_path

    def get_latest_screenshot_path(self):
        screenshots = sorted(
            [f for f in os.listdir(self.screenshot_dir) if f.startswith('screenshot_')],
            key=lambda x: os.path.getmtime(os.path.join(self.screenshot_dir, x)),
            reverse=True
        )
        if screenshots:
            return os.path.join(self.screenshot_dir, screenshots[0])
        return None

    def create_prompt_vars(self, current_action_description, next_goal):
        prompt = {
            "current_action_description": current_action_description,
            "next_goal": next_goal,
        }
        return prompt

    def send_screenshot_to_llm(self, screenshot_path, current_action_description="", next_goal=""):
        with open(screenshot_path, "rb") as image_file:
            encoded_image = base64.b64encode(image_file.read()).decode('utf-8')
        
        prompt_vars = self.create_prompt(current_action_description, next_goal)
        response = self.send_prompt_to_llm(prompt_vars, encoded_image)  # Sending prompt and image separately
        return json.loads(response)

    def send_prompt_to_llm(self, prompt_vars, screenshot_base64):
        raise NotImplementedError("This method should be implemented by subclasses: `def send_prompt_to_llm(self, prompt_vars, screenshot_base64)`")

    def close(self):
        self.save_cookies()
        self.browser.close()
        self.playwright.stop()

    def execute_instructions(self, instructions):
        for instruction in instructions:
            action = instruction['action']
            if action == 'navigate':
                self.navigate(instruction['url'])
            elif action == 'click':
                self.click(instruction['selector'])
            elif action == 'scroll':
                self.scroll(instruction.get('direction', 'down'), instruction.get('amount', 1))
            elif action == 'type':
                self.type_text(instruction['selector'], instruction['text'])
            screenshot_path = self.take_screenshot()
            new_instructions = self.send_screenshot_to_llm(screenshot_path, instruction.get('description', ''), instruction.get('next_goal', ''))
            if new_instructions:
                self.execute_instructions(new_instructions)




from .init import init_vertex, init_genai
from .memory_tools import get_vertex_memories, print_grounding_response, get_google_search_grounding
from .safety import vertex_safety, genai_safety
from .extensions_class import VertexAIExtensions



# https://cloud.google.com/vertex-ai/generative-ai/docs/extensions/create-extension
# https://cloud.google.com/vertex-ai/generative-ai/docs/model-reference/extension#python
from vertexai.preview import extensions
from .init import init_vertex
from ..logging import log
from ..utils.gcp_project import get_gcp_project
from ..utils.parsers import validate_extension_id

# https://github.com/GoogleCloudPlatform/applied-ai-engineering-samples/blob/main/genai-on-vertex-ai/vertex_ai_extensions/notebooks/pandas_code_interpreter.ipynb
import base64
import json
import pprint
import pandas
from io import StringIO

global CODE_INTERPRETER_WRITTEN_FILES
CODE_INTERPRETER_WRITTEN_FILES = []

def get_extension_import_config(
        display_name: str, 
        description: str,
        api_spec_gcs: dict,
        service_account_name: dict,
        tool_use_examples: list):

    tool_use_examples = [
        {
            "extensionOperation": {
                "operationId": "say_hello",
            },
            "displayName": "Say hello in the requested language",
            "query": "Say hello in French",
            "requestParams": {
                "fields": [
                {
                    "key": "apiServicePrompt",
                    "value": {
                    "string_value": "French",
                    }
                }
                ]
            },
            "responseParams": {
                "fields": [
                {
                    "key": "apiServiceOutput",
                    "value": {
                    "string_value": "bonjour",
                    },
                }
                ],
            },
            "responseSummary": "Bonjour"
            }
    ]

    
    return {
        "displayName":  display_name,
        "description": description,
        "manifest": {
            "name": "EXTENSION_NAME_LLM",
            "description": "DESCRIPTION_LLM",
            "apiSpec": {
                "openApiGcsUri": api_spec_gcs,
            },
            "authConfig": {
                "authType": "OAUTH",
                "oauthConfig": {"service_account": service_account_name}
            }
        },
        "toolUseExamples": tool_use_examples, 
        }

# once an extension is available, call it in code here
def create_extension_instance(
        display_name: str,
        description: str,
        open_api_gcs_uri: str,
        llm_name: str=None,
        llm_description: str=None,
        runtime_config: dict=None,
        service_account: str=None,
):
    """
    Args:
    - display_name: for the human. parsed to be used as extension_name
    - description: for the human
    - open_api_gcs_uri: location on GCS where open_ai yaml spec is
    - llm_name: for the model.  If None, uses display_name
    - llm_description: for the model.  If None, uses description
    - service_account: If not specified, the Vertex AI Extension Service Agent is used to execute the extension.
    
    """
    project_id = get_gcp_project()
    extension_name = f"projects/{project_id}/locations/us-central1/extensions/{validate_extension_id(display_name)}"

    extension = extensions.Extension.create(
        extension_name=extension_name,
        display_name=display_name,
        description=description,
        runtime_config=runtime_config or None,
        manifest={
            "name": llm_name or display_name,
            "description": llm_description or description,
            "api_spec": {
                "open_api_gcs_uri": open_api_gcs_uri
            },
            "auth_config": {
                "auth_type": "GOOGLE_SERVICE_ACCOUNT_AUTH",
                "google_service_account_config": service_account or {},
            },
        },
    )
    log.info(f"Created Vertex Extension: {extension_name}")

    return extension



def create_extension_code_interpreter(
        code_artifacts_bucket=None
):

    # only us-central for now
    location = "us-central1"
    init_vertex(location=location)

    runtime_config=None
    if code_artifacts_bucket:
        runtime_config = {"codeInterpreterRuntimeConfig": 
                            {
                                "fileInputGcsBucket": code_artifacts_bucket,
                                "fileOutputGcsBucket": code_artifacts_bucket
                            }
                        }

    llm_description="""
Tool to generate and execute valid Python code from a natural
  language description, or to execute custom Python code.
  Use this tool to:
  - generate and/or execute code for various tasks:
    - perform a wide variety of mathematical calculations, for example, add,
      subtract, multiply, divide, average, power, factorial, quotient,
      formulae, logarithms, random numbers, trigonometric functions, and
      equations;
    - sort, filter, select top results, and otherwise analyze data (including
      data acquired from other tools and Extensions);
    - create visualizations, plot charts, draw graphs, shapes, print results,
      etc.
  - execute custom code and get results and output files.
"""

    code_extension = create_extension_instance(
        display_name="Code Interpreter",
        description="This extension generates and executes code in the specified language",
        open_api_gcs_uri="gs://vertex-extension-public/code_interpreter.yaml",
        llm_name="code_interpreter_tool",
        llm_description=llm_description,
        runtime_config=runtime_config
        )
    log.info(f"Created code extension: {code_extension=}")

    return code_extension

def execute_extension(operation_id: str,
                      operation_params: dict,
                      extension_id: str):

    # only us-central for now
    location = "us-central1"
    init_vertex(location=location)

    if not extension_id.startswith("projects/"):
        project_id=get_gcp_project()
        extension_name = f"projects/{project_id}/locations/{location}/extensions/{extension_id}"
    else:
        extension_name=extension_id
    
    extension = extensions.Extension(extension_name)

    response = extension.execute(
        operation_id=operation_id,
        # {"query": "find the max value in the list: [1,2,3,4,-5]"}
        operation_params=operation_params,
    )

    return response

def execute_code_extension(query:str, filenames: list[str]=None, gcs_files: list[str]=None):

    if filenames and gcs_files:
        raise ValueError("Can't specify both filenames and gcs_files")
    
    extension_code_interpreter = extensions.Extension.from_hub("code_interpreter")

    file_arr=None
    if filenames:
        file_arr = [
            {
                "name": filename,
                "contents":  base64.b64encode(open(filename, "rb").read()).decode()
            }
            for filename in filenames
        ]

    response = extension_code_interpreter.execute(
        operation_id = "generate_and_execute",
        operation_params={
            "query": query,
            "files": file_arr,
            "file_gcs_uris": gcs_files
        })
    
    CODE_INTERPRETER_WRITTEN_FILES.extend(
        [item['name'] for item in response['output_files']])

    if response.get('execution_error'):
        log.error(f"Code Execution Response failed with: {response.get('execution_error')} - maybe retry?")
    
    return response

css_styles = """
<style>
.main_summary {
  font-weight: bold;
  font-size: 14px; color: #4285F4;
  background-color:rgba(221, 221, 221, 0.5); padding:8px;}
</style>
        """

# Parser to visualise the content of returned files as HTML.
def parse_files_to_html(outputFiles, save_files_locally = True):
    IMAGE_FILE_EXTENSIONS = set(["jpg", "jpeg", "png"])
    file_list = []
    details_tml = """<details><summary>{name}</summary><div>{html_content}</div></details>"""

    if not outputFiles:
      return "No Files generated from the code"
    # Sort output_files so images are displayed before other files such as JSON.
    for output_file in sorted(
        outputFiles,
        key=lambda x: x["name"].split(".")[-1] not in IMAGE_FILE_EXTENSIONS,
    ):
        file_name = output_file.get("name")
        file_contents = base64.b64decode(output_file.get("contents"))
        if save_files_locally:
          open(file_name,"wb").write(file_contents)

        if file_name.split(".")[-1] in IMAGE_FILE_EXTENSIONS:
            # Render Image
            file_html_content = ('<img src="data:image/png;base64, '
                                f'{output_file.get("contents")}" />')
        elif file_name.endswith(".json"):
            # Pretty print JSON
            json_pp = pprint.pformat(
                        json.loads(file_contents.decode()),
                        compact=False,
                        width=160)
            file_html_content =  (f'<span>{json_pp}</span>')
        elif file_name.endswith(".csv"):
            # CSV
            csv_md = pandas.read_csv(
                  StringIO(file_contents.decode())).to_markdown(index=False)
            file_html_content = f'<span>{csv_md}</span>'
        elif file_name.endswith(".pkl"):
            # PKL
            file_html_content = f'<span>Preview N/A</span>'
        else:
            file_html_content = f"<span>{file_contents.decode()}</span>"

        file_list.append({'name': file_name, "html_content": file_html_content})

    buffer_html = [ details_tml.format(**_file) for _file in file_list ]
    return "".join(buffer_html)

# Processing code interpreter response to html visualization.
def process_response(response: dict, save_files_locally = None) -> None:

    result_template = """
    <details open>
        <summary class='main_summary'>{summary}:</summary>
        <div><pre>{content}</pre></div>
    </details>
    """

    result = ""
    code = response.get('generated_code')
    if 'execution_result' in response and response['execution_result']!="":
        result = result_template.format(
            summary="Executed Code Output",
            content=response.get('execution_result'))
    else:
        result = result_template.format(
        summary="Executed Code Output",
        content="Code does not produce printable output.")

    if response.get('execution_error', None):
        result += result_template.format(
            summary="Generated Code Raised a (Possibly Non-Fatal) Exception",
            content=response.get('execution_error', None))

    result += result_template.format(
        summary="Files Created <u>(Click on filename to view content)</u>",
        content=parse_files_to_html(
            response.get('output_files', []),
            save_files_locally = True))

    html_content = f"""
{css_styles}
<div id='main'>
    <div id="right">
      <h3>Generated Code by Code Interpreter</h3>
      <pre><code>{code}</code></pre>
    </div>
    <div id="left">
      <h3>Code Execution Results</h3>
      {result}
    </div>
</div>
"""
    if save_files_locally:
        # write to local file
        pass
  
    return html_content



import requests
from ..logging import log
from ..auth import get_header

# Global cache for storing input schemas
langserve_input_schema_cache = {}

def fetch_input_schema(endpoint, vector_name):
    """
    Fetch the input schema from the Langserve endpoint and vector name, including a caching mechanism
    to avoid redundant network calls. If the schema for a given endpoint is already cached, it retrieves
    it directly from the cache; otherwise, it makes an HTTP GET request to fetch the schema.

    Parameters:
    - endpoint (str): The URL of the endpoint from which the schema needs to be fetched.
    - vector_name (str): The name of the vector that might modify the headers sent with the request.

    Returns:
    dict or None: Returns the fetched schema as a dictionary if the request is successful and the schema is valid. Returns None if there is an error fetching the schema.

    This function also logs the fetching process, providing information about the fetched schema or any errors encountered.

    Examples

    ```python
    endpoint = "http://api.example.com/schema/weather"
    vector_name = "weatherQuery"
    schema = fetch_input_schema(endpoint, vector_name)
    if schema:
        print("Schema fetched successfully:", schema)
    else:
        print("Failed to fetch schema")

    cached_schema = fetch_input_schema(endpoint, vector_name)
    print("Cached Schema:", cached_schema)
    ```

    """
    # Check if the schema is already in the cache
    if endpoint in langserve_input_schema_cache:
        return langserve_input_schema_cache[endpoint]

    header = get_header(vector_name)

    try:
        response = requests.get(endpoint, headers = header)
        response.raise_for_status()
        schema = response.json()
        log.info(f"Fetched schema: {schema}")
        # Cache the fetched schema
        langserve_input_schema_cache[endpoint] = schema
        return schema
    except requests.RequestException as e:
        log.error(f"Error fetching input schema: {e}")
        return None

def prepare_request_data(user_input, endpoint, vector_name, **kwargs):
    """
    Prepare the request data for an API call to a Langserve endpoint based on the input schema from the specified endpoint. 
    The function constructs a request payload incorporating user input, endpoint-specific configurations, 
    and additional parameters passed through keyword arguments (kwargs).

    Parameters:
    - user_input (str): The main user input data to be processed.
    - endpoint (str): The endpoint URL or identifier used to fetch the corresponding input schema.
    - vector_name (str): The name of the vector, indicating the specific processing or analysis vector to be used.
    - `**kwargs`: Arbitrary keyword arguments. Special handling for 'configurable' which, if present, 
                is moved to a separate 'config' dictionary in the output payload.

    The function extracts the input schema based on the provided endpoint and vector name, logs this schema, 
    and constructs the payload under the 'input' key based on this schema. If 'configurable' is present in kwargs, 
    it is placed in a separate 'config' dictionary within the payload, allowing configuration settings to be 
    specified separately from the input data.

    Returns:
    dict: A dictionary structured as `{ "input": {...}, "config": {...} }` if 'configurable' is provided; otherwise, the 'config' key is omitted.

    If the input schema is not found or invalid, an error is logged and None is returned.
    """
    input_schema = fetch_input_schema(endpoint, vector_name)
    log.info(f"Found input schema: {input_schema}")
    if input_schema is not None and 'properties' in input_schema:
        key = next(iter(input_schema['properties']))
        input_data = {key: user_input}

        # Check for the special 'configurable' key in kwargs
        config_data = {}
        if 'configurable' in kwargs:
            config_data['configurable'] = kwargs.pop('configurable')

        # Merge kwargs into request_data
        input_data.update(kwargs)
        input_data['vector_name'] = vector_name
        request_data = {"input": input_data}
        if config_data:
            request_data['config'] = config_data

        return request_data
    else:
        log.error("Invalid or no input schema available.")
        return None




try:
    from vertexai.preview import extensions
except ImportError:
    extensions = None

from .init import init_vertex
from ..logging import log
from ..utils.gcp_project import get_gcp_project
from ..utils.parsers import validate_extension_id
import base64
import json
from io import StringIO

class VertexAIExtensions:
    def __init__(self):
        if extensions is None:
            raise ImportError("VertexAIExtensions needs vertexai.previewextensions to be installed. Install via `pip install sunholo[gcp]`")
        
        self.CODE_INTERPRETER_WRITTEN_FILES = []
        self.css_styles = """
        <style>
        .main_summary {
          font-weight: bold;
          font-size: 14px; color: #4285F4;
          background-color:rgba(221, 221, 221, 0.5); padding:8px;}
        </style>
        """
        self.IMAGE_FILE_EXTENSIONS = set(["jpg", "jpeg", "png"])
        self.location = "us-central1"

    def list_extensions(self):
        the_list = extensions.Extension.list()
        
        extensions_list = []
        for ext in the_list:
            extensions_list.append({
                "resource_name": getattr(ext, 'resource_name', ''),
                "display_name": getattr(ext, 'display_name', 'N/A'),
                "description": getattr(ext, 'description', 'N/A')
            })
        
        return extensions_list
        

    def get_extension_import_config(self, display_name: str, description: str,
                                    api_spec_gcs: dict, service_account_name: dict, tool_use_examples: list):
        tool_use_examples = [
            {
                "extensionOperation": {
                    "operationId": "say_hello",
                },
                "displayName": "Say hello in the requested language",
                "query": "Say hello in French",
                "requestParams": {
                    "fields": [
                        {
                            "key": "apiServicePrompt",
                            "value": {
                                "string_value": "French",
                            }
                        }
                    ]
                },
                "responseParams": {
                    "fields": [
                        {
                            "key": "apiServiceOutput",
                            "value": {
                                "string_value": "bonjour",
                            },
                        }
                    ],
                },
                "responseSummary": "Bonjour"
            }
        ]

        return {
            "displayName": display_name,
            "description": description,
            "manifest": {
                "name": "EXTENSION_NAME_LLM",
                "description": "DESCRIPTION_LLM",
                "apiSpec": {
                    "openApiGcsUri": api_spec_gcs,
                },
                "authConfig": {
                    "authType": "OAUTH",
                    "oauthConfig": {"service_account": service_account_name}
                }
            },
            "toolUseExamples": tool_use_examples,
        }

    def create_extension_instance(self, display_name: str, description: str, open_api_gcs_uri: str,
                                  llm_name: str = None, llm_description: str = None, runtime_config: dict = None, service_account: str = None):
        project_id = get_gcp_project()
        extension_name = f"projects/{project_id}/locations/us-central1/extensions/{validate_extension_id(display_name)}"

        extension = extensions.Extension.create(
            extension_name=extension_name,
            display_name=display_name,
            description=description,
            runtime_config=runtime_config or None,
            manifest={
                "name": llm_name or display_name,
                "description": llm_description or description,
                "api_spec": {
                    "open_api_gcs_uri": open_api_gcs_uri
                },
                "auth_config": {
                    "auth_type": "GOOGLE_SERVICE_ACCOUNT_AUTH",
                    "google_service_account_config": service_account or {},
                },
            },
        )
        log.info(f"Created Vertex Extension: {extension_name}")

        return extension

    def execute_extension(self, operation_id: str, operation_params: dict, extension_id: str):
        init_vertex(location=self.location)

        if not extension_id.startswith("projects/"):
            project_id = get_gcp_project()
            extension_name = f"projects/{project_id}/locations/{self.location}/extensions/{extension_id}"
        else:
            extension_name = extension_id

        extension = extensions.Extension(extension_name)

        response = extension.execute(
            operation_id=operation_id,
            operation_params=operation_params,
        )

        return response

    def execute_code_extension(self, 
                               query: str, 
                               filenames: list[str] = None, 
                               gcs_files: list[str] = None,
                               bucket_name: str = None):
        if filenames and gcs_files:
            raise ValueError("Can't specify both filenames and gcs_files")
        
        listed_extensions = self.list_extensions()
        code_interpreter_exists = False
        for ext in listed_extensions:
            if ext.get('display_name') == 'Code Interpreter':
                code_interpreter_exists = True
                extension_code_interpreter = extensions.Extension(ext['resource_name'])
                break

        if not code_interpreter_exists:
            if bucket_name:
                runtime_config = {
                    "codeInterpreterRuntimeConfig": {
                        "fileInputGcsBucket": bucket_name,
                        "fileOutputGcsBucket": bucket_name,
                    }
                }
                log.info(f"Creating buckets with {runtime_config=}")
            else:
                runtime_config = {}

            extension_code_interpreter = extensions.Extension.from_hub("code_interpreter", runtime_config=runtime_config)

        # This field is only applicable when `file_output_gcs_bucket` is specified in `Extension.CodeInterpreterRuntimeConfig`.

        log.info(f"extension_code_interpreter: {extension_code_interpreter.resource_name}")
        operation_params = {"query": query}

        file_arr = None
        if filenames:
            file_arr = [
                {
                    "name": filename,
                    "contents": base64.b64encode(open(filename, "rb").read()).decode()
                }
                for filename in filenames
            ]
            operation_params["files"] = file_arr
        
        if gcs_files:
            operation_params["file_gcs_uris"] = gcs_files
        log.info(f"Executing code interpreter with {operation_params=}")
        response = extension_code_interpreter.execute(
            operation_id="generate_and_execute",
            operation_params=operation_params)

        if response.get('execution_error'):
            #TODO: setup iteration many times with a timeout
            log.error(f"Code Execution Response failed with: {response.get('execution_error')} - maybe retry?")
            new_query = f"""
<original_query>{query}</original_query>
<original_output>{response.get('generated_code')}</original_output>
The code above failed with this error:
<code_error>{response.get('execution_error')}</code_error>
Please try again again to satisfy the original query.
"""
            operation_params = {"query": new_query}
            response = extension_code_interpreter.execute(
                operation_id="generate_and_execute",
                operation_params=operation_params)
            
            if response.get('execution_error'):
                log.error(f"Code Execution Response failed twice: {response.get('execution_error')}")

        if 'output_files' in response:
            self.CODE_INTERPRETER_WRITTEN_FILES.extend(
                [item['name'] for item in response['output_files']])
        
        if 'output_gcs_uris' in response:
            self.CODE_INTERPRETER_WRITTEN_FILES.extend(response['output_gcs_uris'])   

        return response

    def parse_files_to_html(self, outputFiles, save_files_locally=True):
        file_list = []
        details_tml = """<details><summary>{name}</summary><div>{html_content}</div></details>"""

        if not outputFiles:
            return "No Files generated from the code"
        # Sort output_files so images are displayed before other files such as JSON.
        for output_file in sorted(
                outputFiles,
                key=lambda x: x["name"].split(".")[-1] not in self.IMAGE_FILE_EXTENSIONS,
        ):
            file_name = output_file.get("name")
            file_contents = base64.b64decode(output_file.get("contents"))
            if save_files_locally:
                open(file_name, "wb").write(file_contents)

            if file_name.split(".")[-1] in self.IMAGE_FILE_EXTENSIONS:
                # Render Image
                file_html_content = ('<img src="data:image/png;base64, '
                                     f'{output_file.get("contents")}" />')
            elif file_name.endswith(".json"):
                import pprint
                # Pretty print JSON
                json_pp = pprint.pformat(
                    json.loads(file_contents.decode()),
                    compact=False,
                    width=160)
                file_html_content = (f'<span>{json_pp}</span>')
            elif file_name.endswith(".csv"):
                # CSV
                try:
                    import pandas
                except ImportError:
                    log.error("Need pandas for csv processing")
                csv_md = pandas.read_csv(
                    StringIO(file_contents.decode())).to_markdown(index=False)
                file_html_content = f'<span>{csv_md}</span>'
            elif file_name.endswith(".pkl"):
                # PKL
                file_html_content = f'<span>Preview N/A</span>'
            else:
                file_html_content = f"<span>{file_contents.decode()}</span>"

            file_list.append({'name': file_name, "html_content": file_html_content})

        buffer_html = [details_tml.format(**_file) for _file in file_list]
        return "".join(buffer_html)
    
    def process_response(self, response: dict, save_file_name=None) -> str:
        result_template = """
        <details open>
            <summary class='main_summary'>{summary}:</summary>
            <div><pre>{content}</pre></div>
        </details>
        """

        result = ""
        code = response.get('generated_code')
        if 'execution_result' in response and response['execution_result'] != "":
            result = result_template.format(
                summary="Executed Code Output",
                content=response.get('execution_result'))
        else:
            result = result_template.format(
                summary="Executed Code Output",
                content="Code does not produce printable output.")

        if response.get('execution_error', None):
            result += result_template.format(
                summary="Generated Code Raised a (Possibly Non-Fatal) Exception",
                content=response.get('execution_error', None))

        result += result_template.format(
            summary="Files Created <u>(Click on filename to view content)</u>",
            content=self.parse_files_to_html(
                response.get('output_files', []),
                save_files_locally=True))

        html_content = f"""
{self.css_styles}
<div id='main'>
    <h3>Generated Code by Code Interpreter</h3>
    <pre><code>{code}</code></pre>
    <h3>Code Execution Results</h3>
    {result}
</div>
"""
        if save_file_name:
            with open(save_file_name or 'code_execution_results.html', 'w') as file:
                file.write(html_content)

        return html_content


# from https://github.com/ray-project/docu-mentor
import base64
import httpx
from dotenv import load_dotenv
import jwt
import os
import time

load_dotenv()



APP_ID = os.environ.get("APP_ID")
PRIVATE_KEY = os.environ.get("PRIVATE_KEY", "")

# with open('private-key.pem', 'r') as f:
#     PRIVATE_KEY = f.read()

def generate_jwt():
    payload = {
        "iat": int(time.time()),
        "exp": int(time.time()) + (10 * 60),
        "iss": APP_ID,
    }
    if PRIVATE_KEY:
        jwt_token = jwt.encode(payload, PRIVATE_KEY, algorithm="RS256")
        return jwt_token
    raise ValueError("PRIVATE_KEY not found.")


async def get_installation_access_token(jwt, installation_id):
    url = f"https://api.github.com/app/installations/{installation_id}/access_tokens"
    headers = {
        "Authorization": f"Bearer {jwt}",
        "Accept": "application/vnd.github.v3+json",
    }
    async with httpx.AsyncClient() as client:
        response = await client.post(url, headers=headers)
        return response.json()["token"]


def get_diff_url(pr):
    """GitHub 302s to this URL."""
    original_url = pr.get("url")
    parts = original_url.split("/")
    owner, repo, pr_number = parts[-4], parts[-3], parts[-1]
    return f"https://patch-diff.githubusercontent.com/raw/{owner}/{repo}/pull/{pr_number}.diff"


async def get_branch_files(pr, branch, headers):
    original_url = pr.get("url")
    parts = original_url.split("/")
    owner, repo = parts[-4], parts[-3]
    url = f"https://api.github.com/repos/{owner}/{repo}/git/trees/{branch}?recursive=1"
    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers)
        tree = response.json().get('tree', [])
        files = {}
        for item in tree:
            if item['type'] == 'blob':
                file_url = item['url']
                print(file_url)
                file_response = await client.get(file_url, headers=headers)
                content = file_response.json().get('content', '')
                # Decode the base64 content
                decoded_content = base64.b64decode(content).decode('utf-8')
                files[item['path']] = decoded_content
        return files


async def get_pr_head_branch(pr, headers):
    original_url = pr.get("url")
    parts = original_url.split("/")
    owner, repo, pr_number = parts[-4], parts[-3], parts[-1]
    url = f"https://api.github.com/repos/{owner}/{repo}/pulls/{pr_number}"

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers)

        # Check if the response is successful
        if response.status_code != 200:
            print(f"Error: Received status code {response.status_code}")
            print("Response body:", response.text)
            return ''

        # Safely get the 'ref'
        data = response.json()
        head_data = data.get('head', {})
        ref = head_data.get('ref', '')
        return ref


def files_to_diff_dict(diff):
    files_with_diff = {}
    current_file = None
    for line in diff.split("\n"):
        if line.startswith("diff --git"):
            current_file = line.split(" ")[2][2:]
            files_with_diff[current_file] = {"text": []}
        elif line.startswith("+") and not line.startswith("+++"):
            files_with_diff[current_file]["text"].append(line[1:])
    return files_with_diff


def parse_diff_to_line_numbers(diff):
    files_with_line_numbers = {}
    current_file = None
    line_number = 0
    for line in diff.split("\n"):
        if line.startswith("diff --git"):
            current_file = line.split(" ")[2][2:]
            files_with_line_numbers[current_file] = []
            line_number = 0
        elif line.startswith("@@"):
            line_number = int(line.split(" ")[2].split(",")[0][1:]) - 1
        elif line.startswith("+") and not line.startswith("+++"):
            files_with_line_numbers[current_file].append(line_number)
            line_number += 1
        elif not line.startswith("-"):
            line_number += 1
    return files_with_line_numbers


def get_context_from_files(files, files_with_line_numbers, context_lines=2):
    context_data = {}
    for file, lines in files_with_line_numbers.items():
        file_content = files[file].split("\n")
        context_data[file] = []
        for line in lines:
            start = max(line - context_lines, 0)
            end = min(line + context_lines + 1, len(file_content))
            context_data[file].append('\n'.join(file_content[start:end]))
    return context_data

app = FastAPI()


async def handle_webhook(request: Request):
    data = await request.json()

    installation = data.get("installation")
    if installation and installation.get("id"):
        installation_id = installation.get("id")
        logger.info(f"Installation ID: {installation_id}")

        JWT_TOKEN = generate_jwt()

        installation_access_token = await get_installation_access_token(
            JWT_TOKEN, installation_id
        )

        headers = {
            "Authorization": f"token {installation_access_token}",
            "User-Agent": "docu-mentor-bot",
            "Accept": "application/vnd.github.VERSION.diff",
        }
    else:
        raise ValueError("No app installation found.")

    # If PR exists and is opened
    if "pull_request" in data.keys() and (
        data["action"] in ["opened", "reopened"]
    ):  # use "synchronize" for tracking new commits
        pr = data.get("pull_request")

        # Greet the user and show instructions.
        async with httpx.AsyncClient() as client:
            await client.post(
                f"{pr['issue_url']}/comments",
                json={"body": GREETING},
                headers=headers,
            )
        return JSONResponse(content={}, status_code=200)

    # Check if the event is a new or modified issue comment
    if "issue" in data.keys() and data.get("action") in ["created", "edited"]:
        issue = data["issue"]

        # Check if the issue is a pull request
        if "/pull/" in issue["html_url"]:
            pr = issue.get("pull_request")

            # Get the comment body
            comment = data.get("comment")
            comment_body = comment.get("body")
            # Remove all whitespace characters except for regular spaces
            comment_body = comment_body.translate(
                str.maketrans("", "", string.whitespace.replace(" ", ""))
            )

            # Skip if the bot talks about itself
            author_handle = comment["user"]["login"]

            # Check if the bot is mentioned in the comment
            if (
                author_handle != "docu-mentor[bot]"
                and "@docu-mentor run" in comment_body
            ):
                async with httpx.AsyncClient() as client:
                    # Fetch diff from GitHub
                    files_to_keep = comment_body.replace(
                        "@docu-mentor run", ""
                    ).split(" ")
                    files_to_keep = [item for item in files_to_keep if item]

                    logger.info(files_to_keep)

                    url = get_diff_url(pr)
                    diff_response = await client.get(url, headers=headers)
                    diff = diff_response.text

                    files_with_lines = parse_diff_to_line_numbers(diff)

                    # Get head branch of the PR
                    headers["Accept"] = "application/vnd.github.full+json"
                    head_branch = await get_pr_head_branch(pr, headers)

                    # Get files from head branch
                    head_branch_files = await get_branch_files(pr, head_branch, headers)
                    print("HEAD FILES", head_branch_files)

                    # Enrich diff data with context from the head branch.
                    context_files = get_context_from_files(head_branch_files, files_with_lines)

                    # Filter the dictionary
                    if files_to_keep:
                        context_files = {
                            k: context_files[k]
                            for k in context_files
                            if any(sub in k for sub in files_to_keep)
                        }

                    # Get suggestions from Docu Mentor
                    content, model, prompt_tokens, completion_tokens = \
                        ray_mentor(context_files) if ray.is_initialized() else mentor(context_files)


                    # Let's comment on the PR
                    await client.post(
                        f"{comment['issue_url']}/comments",
                        json={
                            "body": f":rocket: Docu Mentor finished "
                            + "analysing your PR! :rocket:\n\n"
                            + "Take a look at your results:\n"
                            + f"{content}\n\n"
                            + "This bot is  powered by "
                            + "[Sunholo Multivac](https://www.sunholo.com/).\n"
                            + f"It used the model {model}, used {prompt_tokens} prompt tokens, "
                            + f"and {completion_tokens} completion tokens in total."
                        },
                        headers=headers,
                    )

@serve.deployment(route_prefix="/")
@serve.ingress(app)
class ServeBot:
    @app.get("/")
    async def root(self):
        return {"message": "Docu Mentor reporting for duty!"}

    @app.post("/webhook/")
    async def handle_webhook_route(self, request: Request):
        return await handle_webhook(request)




#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import log
from ..utils import load_config_key, load_config

def route_vac(vector_name: str) -> str :
    """
    Considers what VAC this vector_name belongs to
    """
    agent_url = load_config_key('agent_url', vector_name=vector_name, kind="vacConfig")
    if agent_url:
        log.info('agent_url found in llm_config.yaml')
        return agent_url

    agent = load_config_key('agent', vector_name, kind="vacConfig")
    log.info(f'agent_type: {agent}')

    agent_route, _ = load_config('config/cloud_run_urls.json')
    log.info(f'agent_route: {agent_route}')

    try:
        agent_url = agent_route[agent]
    except KeyError:
        raise ValueError(f'agent_url not found for {agent}')
    
    log.info(f'agent_url: {agent_url}')
    return agent_url

def route_endpoint(vector_name, method = 'post', override_endpoint=None):

    agent_type = load_config_key('agent_type', vector_name, kind="vacConfig")
    if not agent_type:
        agent_type = load_config_key('agent', vector_name, kind="vacConfig")

    stem = route_vac(vector_name) if not override_endpoint else override_endpoint
    
    agents_config = load_config_key(agent_type, vector_name, kind="agentConfig")
    
    log.info(f"agents_config: {agents_config}")
    if method not in agents_config:
        raise ValueError(f"Invalid method '{method}' for agent configuration.")

    # 'post' or 'get'
    endpoints_config = agents_config[method]

    log.info(f"endpoints_config: {endpoints_config}")
    # Replace placeholders in the config
    endpoints = {}
    for key, value in endpoints_config.items():
        format_args = {'stem': stem}
        if '{vector_name}' in value and vector_name is not None:
            format_args['vector_name'] = vector_name
        
        if not isinstance(value, str):
            log.warning('endpoint value not string? format_args: {format_args} - value: {value} - key: {key}')
            
        endpoints[key] = value.format(**format_args)

    return endpoints




import pytest
from unittest.mock import patch, MagicMock
from sunholo.chunker.data_to_embed_pubsub import data_to_embed_pubsub

# Mock external calls within the function
@patch('sunholo.chunker.data_to_embed_pubsub.process_pubsub_message', return_value=({}, {}, 'test_vector'))
@patch('sunholo.chunker.data_to_embed_pubsub.process_chunker_data', return_value='processed_data')
def test_data_to_embed_pubsub(mock_process_chunker_data, mock_process_pubsub_message):
    # Test the function with various inputs including edge cases
    assert data_to_embed_pubsub({}) == 'processed_data'
    assert data_to_embed_pubsub({'key': 'value'}) == 'processed_data'
    mock_process_pubsub_message.assert_called()
    mock_process_chunker_data.assert_called()

    # Ensure tests are self-contained and do not require external dependencies
    mock_process_pubsub_message = MagicMock(return_value=({}, {}, 'test_vector'))
    mock_process_chunker_data = MagicMock(return_value='processed_data')
    assert data_to_embed_pubsub({'key': 'value'}) == 'processed_data'

    # Validate the function's output against expected results
    expected_output = 'processed_data'
    actual_output = data_to_embed_pubsub({'key': 'value'})
    assert actual_output == expected_output, f"Expected {expected_output}, got {actual_output}"



# from https://github.com/sunholo-data/genai-databases-retrieval-app/blob/main/langchain_tools_demo/agent.py
import inspect

from typing import Dict, Optional
from ..utils.config import load_config_key, load_config
from ..utils.gcp import is_running_on_cloudrun
from ..utils.api_key import has_multivac_api_key, get_multivac_api_key
from ..logging import log
from ..agents.route import route_vac

def get_run_url(vector_name=None):

    if not vector_name:
        raise ValueError('Vector name was not specified')
    
    cloud_urls = route_vac(vector_name)
    
    cloud_urls, _ = load_config('config/cloud_run_urls.json')
    agent = load_config_key("agent", vector_name=vector_name, kind="vacConfig")

    try:
        log.info(f'Looking up URL for {agent}')
        url = cloud_urls[agent]
        return url
    except KeyError:
        raise ValueError(f'Could not find cloud_run_url for {agent} within {cloud_urls}')

def get_id_token(url: str) -> str:
    """Helper method to generate ID tokens for authenticated requests"""
    # Use Application Default Credentials on Cloud Run
    if is_running_on_cloudrun():
        import google.auth.transport.requests  # type: ignore
        import google.oauth2.id_token  # type: ignore
        auth_req = google.auth.transport.requests.Request()
        log.info(f'Got id_token for {url}')
        return google.oauth2.id_token.fetch_id_token(auth_req, url)
    else:
        # Use gcloud credentials locally
        import subprocess

        return (
            subprocess.run(
                ["gcloud", "auth", "print-identity-token"],
                stdout=subprocess.PIPE,
                check=True,
            )
            .stdout.strip()
            .decode()
        )

def get_header(vector_name) -> Optional[dict]:
    if has_multivac_api_key():
        
        return {"x-api-key": get_multivac_api_key()}

    if is_running_on_cloudrun():
        run_url = get_run_url(vector_name)
    else:
        run_url = "http://127.0.0.1:8080"

    # Append ID Token to make authenticated requests to Cloud Run services
    frame = inspect.currentframe()
    caller_frame = frame.f_back if frame is not None else None  # One level up in the stack
    deets = {
        'message': 'Authenticating for run_url',
        'run_url': run_url
    }
    if caller_frame:
        deets = {
                'message': 'Authenticating for run_url',
                'file': caller_frame.f_code.co_filename,
                'line': str(caller_frame.f_lineno),  
                'function': caller_frame.f_code.co_name,
                'run_url': run_url
            }
    log.info(f"Authenticating for run_url {run_url} from {caller_frame.f_code.co_name}")
    id_token = get_id_token(run_url)
    headers = {"Authorization": f"Bearer {id_token}"}
    #log.info(f"id_token {id_token}")
    return headers


from ..logging import log

def create_lancedb_index(bucket, vector_name, num_partitions=256, num_sub_vectors=96, text_key="text"):
    try:
        import lancedb
        #import tantivy
    except ImportError:
        raise ValueError("Could not import lancedb module, install via `pip install sunholo[database]`")
        
    try:
        db = lancedb.connect(bucket) 
        tbl = db.open_table(vector_name)

        tbl.create_index(num_partitions=num_partitions, num_sub_vectors=num_sub_vectors)
        tbl.create_fts_index(text_key)
        log.info(f'Index creation for {vector_name} success')
    except Exception as e: 
        log.info(f'Index creation for {vector_name} failed: {e}')
    
    



from ..logging import log
from ..utils.gcp_project import get_gcp_project
import os

def init_genai():
    """
    There are some features that come to the google.generativeai first, 
    which needs to be authenticated via a GOOGLE_API_KEY environment variable, 
    created via the Google AI Console at https://aistudio.google.com/app/apikey 
    """
    try:
        import google.generativeai as genai
    except ImportError:
        raise ImportError("google.generativeai not installed, please install via 'pip install sunholo[gcp]")
    
    GOOGLE_API_KEY=os.getenv('GOOGLE_API_KEY')
    if not GOOGLE_API_KEY:
        raise ValueError("google.generativeai needs GOOGLE_API_KEY set in environment variable")

    genai.configure(api_key=GOOGLE_API_KEY)

def init_vertex(gcp_config=None, location="eu"):
    """
    Initializes the Vertex AI environment using the provided Google Cloud Platform configuration.

    This function configures the Vertex AI API session with specified project and location details
    from the gcp_config dictionary. It is essential to call this function at the beginning of a session
    before performing any operations related to Vertex AI.

    Parameters:
        gcp_config (dict): A dictionary containing the Google Cloud Platform configuration with keys:
            - 'project_id': The Google Cloud project ID to configure for Vertex AI.
            - 'location': The Google Cloud region to configure for Vertex AI.
            If default None it will derive it from the environment

    Raises:
        KeyError: If the necessary keys ('project_id' or 'location') are missing in the gcp_config dictionary.
        ModuleNotFoundError: If the Vertex AI module is not installed and needs to be installed via pip.

    Example:
    ```python
    gcp_config = {
         'project_id': 'your-project-id',
         'location': 'us-central1'
    }
    init_vertex(gcp_config)
    # This will initialize the Vertex AI session with the provided project ID and location.

    Note:
        Ensure that the 'vertexai' module is installed and correctly configured before calling this function.
        The function assumes that the required 'vertexai' library is available and that the logging setup is already in place.
    """
    try:
        import vertexai
    except ImportError:
        log.error("Need to install vertexai module via `pip install sunholo[gcp]`")

        return None
    
    if gcp_config:
        # Initialize Vertex AI API once per session
        project_id = gcp_config.get('project_id')
        location = gcp_config.get('location') or location
    else:
        project_id = get_gcp_project()

    vertexai.init(project=project_id, location=location)
